@@ -29,6 +29,8 @@ module Control.Alternative.Free
 
 import Control.Applicative
 import Data.Functor.Apply
+import Data.Functor.Alt ((<!>))
+import qualified Data.Functor.Alt as Alt
 import Data.Semigroup
 import Data.Typeable
 
@@ -98,6 +100,10 @@ instance (Functor f) => Apply (Alt f) where
   (<.>) = (<*>)
   {-# INLINE (<.>) #-}
 
+instance (Functor f) => Alt.Alt (Alt f) where
+  (<!>) = (<|>)
+  {-# INLINE (<!>) #-}
+
 instance (Functor f) => Alternative (Alt f) where
   empty = Alt []
   {-# INLINE empty #-}
