@@ -6,6 +6,10 @@
 {-# LANGUAGE Rank2Types #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 
+#ifndef MIN_VERSION_base
+#define MIN_VERSION_base(x,y,z) 1
+#endif
+
 #ifndef MIN_VERSION_mtl
 #define MIN_VERSION_mtl(x,y,z) 1
 #endif
@@ -40,6 +44,15 @@ module Control.Monad.Trans.Iter
   -- monad encapsulates errors, the 'Iter' monad encapsulates
   -- non-termination. The 'IterT' transformer generalizes non-termination to any monadic
   -- computation.
+  --
+  -- Computations in 'IterT' (or 'Iter') can be composed in two ways:
+  --
+  -- * /Sequential:/ Using the 'Monad' instance, the result of a computation
+  --   can be fed into the next.
+  --
+  -- * /Parallel:/ Using the 'MonadPlus' instance, several computations can be
+  --   executed concurrently, and the first to finish will prevail.
+  --   See also the <examples/Cabbage.lhs cabbage example>.
 
   -- * The iterative monad transformer
     IterT(..)
@@ -51,6 +64,7 @@ module Control.Monad.Trans.Iter
   , liftIter
   , cutoff
   , never
+  , untilJust
   , interleave, interleave_
   -- * Consuming iterative monads
   , retract
@@ -63,6 +77,7 @@ module Control.Monad.Trans.Iter
   ) where
 
 import Control.Applicative
+import Control.Monad.Catch (MonadCatch(..), MonadThrow(..))
 import Control.Monad (ap, liftM, MonadPlus(..), join)
 import Control.Monad.Fix
 import Control.Monad.Trans.Class
@@ -78,9 +93,7 @@ import Data.Bitraversable
 import Data.Either
 import Data.Functor.Bind hiding (join)
 import Data.Functor.Identity
-import Data.Foldable hiding (fold)
 import Data.Function (on)
-import Data.Traversable hiding (mapM)
 import Data.Monoid
 import Data.Semigroup.Foldable
 import Data.Semigroup.Traversable
@@ -88,6 +101,11 @@ import Data.Typeable
 import Data.Data
 import Prelude.Extras
 
+#if !(MIN_VERSION_base(4,8,0))
+import Data.Foldable hiding (fold)
+import Data.Traversable hiding (mapM)
+#endif
+
 -- | The monad supporting iteration based over a base monad @m@.
 --
 -- @
@@ -158,7 +176,7 @@ instance Monad m => Monad (IterT m) where
   {-# INLINE return #-}
   IterT m >>= k = IterT $ m >>= either (runIterT . k) (return . Right . (>>= k))
   {-# INLINE (>>=) #-}
-  fail = IterT . fail
+  fail _ = never
   {-# INLINE fail #-}
 
 instance Monad m => Apply (IterT m) where
@@ -173,16 +191,19 @@ instance MonadFix m => MonadFix (IterT m) where
   mfix f = IterT $ mfix $ runIterT . f . either id (error "mfix (IterT m): Right")
   {-# INLINE mfix #-}
 
-instance MonadPlus m => Alternative (IterT m) where
-  empty = IterT mzero
+instance Monad m => Alternative (IterT m) where
+  empty = mzero
   {-# INLINE empty #-}
-  IterT a <|> IterT b = IterT (mplus a b)
+  (<|>) = mplus
   {-# INLINE (<|>) #-}
 
-instance MonadPlus m => MonadPlus (IterT m) where
-  mzero = IterT mzero
+-- | Capretta's 'race' combinator. Satisfies left catch.
+instance Monad m => MonadPlus (IterT m) where
+  mzero = never
   {-# INLINE mzero #-}
-  IterT a `mplus` IterT b = IterT (mplus a b)
+  (IterT x) `mplus` (IterT y) = IterT $ x >>= either
+                                (return . Left)
+                                (flip liftM y . second . mplus)
   {-# INLINE mplus #-}
 
 instance MonadTrans IterT where
@@ -256,6 +277,14 @@ instance Monad m => MonadFree Identity (IterT m) where
   wrap = IterT . return . Right . runIdentity
   {-# INLINE wrap #-}
 
+instance MonadThrow m => MonadThrow (IterT m) where
+  throwM = lift . throwM
+  {-# INLINE throwM #-}
+
+instance MonadCatch m => MonadCatch (IterT m) where
+  catch (IterT m) f = IterT $ liftM (fmap (`Control.Monad.Catch.catch` f)) m `Control.Monad.Catch.catch` (runIterT . f)
+  {-# INLINE catch #-}
+
 -- | Adds an extra layer to a free monad value.
 --
 -- In particular, for the iterative monad 'Iter', this makes the
@@ -297,6 +326,20 @@ liftIter = hoistIterT (return . runIdentity)
 never :: (Monad f, MonadFree f m) => m a
 never = delay never
 
+-- | Repeatedly run a computation until it produces a 'Just' value.
+-- This can be useful when paired with a monad that has side effects.
+--
+-- For example, we may have @genId :: IO (Maybe Id)@ that uses a random
+-- number generator to allocate ids, but fails if it finds a collision.
+-- We can repeatedly run this with
+--
+-- @
+-- 'retract' ('untilJust' genId) :: IO Id
+-- @
+untilJust :: (Monad m) => m (Maybe a) -> IterT m a
+untilJust f = maybe (delay (untilJust f)) return =<< lift f
+{-# INLINE untilJust #-}
+
 -- | Cuts off an iterative computation after a given number of
 -- steps. If the number of steps is 0 or less, no computation nor
 -- monadic effects will take place.
@@ -416,6 +459,8 @@ iterDataType = mkDataType "Control.Monad.Iter.IterT" [iterConstr]
 
 {- $examples
 
-<examples/MandelbrotIter.lhs Mandelbrot>
+* <examples/MandelbrotIter.lhs Rendering the Mandelbrot set>
+
+* <examples/Cabbage.lhs The wolf, the sheep and the cabbage>
 
 -}
